__all__ = ('__version__', 'DataModel')

from pbr.version import VersionInfo

from . import patches  # noqa
from .datamodel import DataModel  # noqa

# Check the PBR version module docs for other options than release_string()
__version__ = VersionInfo('flask-restless-datamodel').release_string()



import flask
import pytest
from requests_flask_adapter import Session


@pytest.fixture
def app():
    app = flask.Flask(__name__)
    app.config['DEBUG'] = True
    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///'
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    return app


@pytest.fixture(scope='function')
def client_maker():
    def client(app):
        Session.register('http://app', app)
        return Session()

    return client



import json
from collections import defaultdict
from functools import wraps

import flask_restless
from cereal_lazer import Cereal
from flask import Response, abort
from flask.testing import EnvironBuilder
from pbr.version import VersionInfo

from .helpers import ModelConfiguration
from .render import DataModelRenderer


def catch_model_view(dispatch_request, getaway_car):
    """
    This is the actual point where we catch the relevant configuration made
    by Flask-Restless. Currently we are only interested in the include and
    exclude columns, but as needs may arise in the future this method may
    grow.

    Flask-Restless generates APIView classes on the fly for each registered model
    and uses this class as a view_func. We monkey patch the call to get access
    to the parameters and then return it back to the original method.
    Due to Flask's "as_view" implementation, it is the only entry point to
    retrieve this information without restrictions. There are other ways to
    retrieve it, but it relies on import and initialisation order, and quickly
    becomes dirty and restrictive.

    And this way, we're at least dropping the restrictive part :)
    """

    def wrapper(self, *args, **kwargs):
        # Putting back the old and original dispatch_request method to continue
        # normal operation from this point on.
        self.__class__.dispatch_request = dispatch_request
        getaway_car.append(self)
        return {}

    return wrapper


def attach_listener(create_blueprint, data_model):
    @wraps(create_blueprint)
    def wrapper(model, *args, **kwargs):
        app = kwargs.get('app', data_model.api_manager.app)
        if isinstance(model, DataModel):
            data_model.init(app)
            kwargs['preprocessors'] = data_model.processors
            return create_blueprint(model, *args, **kwargs)
        blueprint = create_blueprint(model, *args, **kwargs)
        app.register_blueprint(blueprint)
        api_info = data_model.api_manager.created_apis_for[model]
        data_model.register_model(model, api_info, app)
        return blueprint

    return wrapper


class DataModel(object):
    __tablename__ = 'flask-restless-datamodel'

    def __init__(self, api_manager, **options):
        """
        In Flask-Restless, it is up to you to choose which models you would like
        to make available as an api. It should also be a choice to expose your
        datamodel, and preferably in the same intuitive way that you register
        your models.

        This object functions as a puppet model to give the user the feeling
        like they are registering just another model they want to expose.
        """
        api_manager.create_api_blueprint = attach_listener(
            api_manager.create_api_blueprint, self)
        self.api_manager = api_manager
        vi = VersionInfo('flask-restless-datamodel')
        serialize_naively = options.get('serialize_naively', False)
        self.data_model = {
            'FlaskRestlessDatamodel': {
                'server_version': vi.release_string(),
                'serialize_naively': serialize_naively
            }
        }
        self.polymorphic_info = defaultdict(dict)
        self.options = options
        self.model_renderer = None
        self.model_views = {}
        self.app = None
        self.cereal = Cereal(
            raise_load_errors=options.get('raise_load_errors', True),
            serialize_naively=serialize_naively)

    def init(self, app):
        db = self.api_manager.flask_sqlalchemy_db
        self.app = app
        if not hasattr(app, 'extensions'):
            app.extensions = {}
        app.extensions['cereal'] = self.cereal
        self.model_renderer = DataModelRenderer(app, db, self.options)
        # render datamodel for models that were already registered to
        # flask-restless
        for model, api_info in self.api_manager.created_apis_for.items():
            self.register_model(model, api_info, app)

    def register_model(self, model, api_info, app):
        name = model.__name__
        view = self.get_restless_view(model, api_info, app)
        blueprint = app.blueprints[api_info.blueprint_name]
        collection_name = api_info.collection_name

        conf = ModelConfiguration(collection_name, view, blueprint)
        render = self.model_renderer.render(model, conf)

        polymorphic_info = self.model_renderer.render_polymorphic(
            model, self.polymorphic_info[name])

        if 'parent' in polymorphic_info:
            parent = polymorphic_info['parent']
            identity = polymorphic_info['identity']
            self.polymorphic_info[parent][identity] = name

        if polymorphic_info:
            render['polymorphic'] = polymorphic_info

        self.data_model[name] = render
        self.app.register_blueprint(blueprint)

    @property
    def processors(self):
        return {
            'GET': [self.intercept_and_return_datamodel],
            'GET_MANY': [self.intercept_and_return_datamodel]
        }

    def intercept_and_return_datamodel(self, *args, **kwargs):
        """
        This method must be called as a preprocessor to the actual restless
        api call. It will construct the json data model, if it hasn't already been
        constructed, and return it.

        The goal of running this method as a preprocessor is so that we have
        chance to intercept the request before it gets sent to do the actual
        restless database query.

        Since this model is not an actual SQLAlchemy model, it will crash when
        actual db queries are executed on it. This is why we're (mis)using the
        flask abort to prematurely break off the normal restless flow,
        as by now we have all the data we need to return our request.
        """
        # (Mis)using the flask abort to return the datamodel before the
        # request gets forwarded to the actual db querying
        abort(
            Response(
                response=json.dumps(self.data_model),
                mimetype='application/json'))

    def get_restless_view(self, model, api_info, app):
        """
        This method will try to find the corresponding view within the registered
        blueprints in flask-restless and momentarily replace it with a function
        that is able to distil the relevant infomation we need to construct a
        datamodel that is conform to what constraints were defined in
        flask-restless when registering models.
        After the first call it will replace the function handle back to its
        original function.
        """
        api_format = flask_restless.APIManager.APINAME_FORMAT
        endpoint = api_format.format('{1}.{0}'.format(*api_info))

        view_func = app.view_functions[endpoint]

        getaway_car = []
        dispatch_fn = catch_model_view(view_func.view_class.dispatch_request,
                                       getaway_car)
        view_func.view_class.dispatch_request = dispatch_fn

        with app.request_context(self.build_stub_environ(app)):
            view_func().json

        view = getaway_car[0]
        return view

    def build_stub_environ(self, app):
        kw = {'base_url': 'http://localhost'}
        builder = EnvironBuilder(self.app, **kw)
        try:
            environ = builder.get_environ()
        finally:
            builder.close()
        return environ



import inspect
import sys

import flask_restless
from packaging import version
from pbr.version import VersionInfo


def primary_key_names(model):
    """Returns all the primary keys for a model."""
    return [
        key for key, field in inspect.getmembers(model)
        if isinstance(field, QueryableAttribute) and hasattr(
            field, 'property') and isinstance(field.property, ColumnProperty)
        and field.property.columns[0].primary_key
    ]


def get_related_model(model, relationname):
    """Gets the class of the model to which `model` is related by the attribute
    whose name is `relationname`.

    """
    if hasattr(model, relationname):
        attr = getattr(model, relationname)
        if hasattr(attr, 'property') \
                and isinstance(attr.property, RelProperty):
            return attr.property.mapper.class_
        if isinstance(attr, ASSOCIATION_PROXIES_KLASSES):
            return flask_restless.helpers.get_related_association_proxy_model(
                attr)
    return None


def get_relations(model):
    """Returns a list of relation names of `model` (as a list of strings)."""

    def is_accepted(k):
        return (not (k.startswith('__')
                     or k in flask_restless.helpers.RELATION_BLACKLIST)
                and not k.startswith("_AssociationProxy")
                and not k.startswith("_ReadOnlyAssociationProxy")
                and get_related_model(model, k))

    return [k for k in dir(model) if is_accepted(k)]


def is_like_list(instance, relation):
    """Returns ``True`` if and only if the relation of `instance` whose name is
    `relation` is list-like.

    A relation may be like a list if, for example, it is a non-lazy one-to-many
    relation, or it is a dynamically loaded one-to-many.

    """
    if relation in instance._sa_class_manager:
        return instance._sa_class_manager[relation].property.uselist
    elif hasattr(instance, relation):
        attr = getattr(instance._sa_instance_state.class_, relation)
        if hasattr(attr, 'property'):
            return attr.property.uselist
    related_value = getattr(type(instance), relation, None)
    if isinstance(related_value, ASSOCIATION_PROXIES_KLASSES):
        local_prop = related_value.local_attr.prop
        if isinstance(local_prop, RelProperty):
            return local_prop.uselist
    return False


def apply_patches():
    needs_patching = (primary_key_names, get_related_model, get_relations,
                      is_like_list)

    for func in needs_patching:
        funcname = func.__name__
        restless_mods = [
            m for m in sys.modules if m.startswith('flask_restless')
        ]
        for mod in restless_mods:
            if funcname in dir(sys.modules[mod]):
                setattr(sys.modules[mod], funcname, func)


sqla_version = version.parse(VersionInfo('sqlalchemy').release_string())
if sqla_version >= version.parse('1.3.0'):
    from sqlalchemy.ext.associationproxy import (
        AssociationProxy, ObjectAssociationProxyInstance)
    from sqlalchemy.orm import (RelationshipProperty as RelProperty,
                                ColumnProperty)
    from sqlalchemy.orm.attributes import QueryableAttribute

    ASSOCIATION_PROXIES_KLASSES = (AssociationProxy,
                                   ObjectAssociationProxyInstance)
    apply_patches()



from datetime import date
from unittest.mock import patch

import flask_restless
import pytest
from cereal_lazer import Cereal
from flask_restless_datamodel import DataModel, __version__
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.hybrid import hybrid_property


def test_datamodel(app, client_maker):
    db = SQLAlchemy(app)

    class Person(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        name = db.Column(db.Unicode, unique=True)
        birth_date = db.Column(db.Date)

    class Computer(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        name = db.Column(db.Unicode, unique=True)
        vendor = db.Column(db.Unicode)
        purchase_time = db.Column(db.DateTime)
        owner_id = db.Column(db.Integer, db.ForeignKey('person.id'))
        owner = db.relationship(
            'Person', backref=db.backref('computers', lazy='dynamic'))
        owner_name = association_proxy('owner', 'name')
        peers = association_proxy('owner', 'computers')

    db.create_all()

    manager = flask_restless.APIManager(app, flask_sqlalchemy_db=db)
    manager.create_api(Person, methods=['GET'], include_columns=['name'])
    data_model = DataModel(manager)
    manager.create_api(data_model, methods=['GET'])
    manager.create_api(
        Computer,
        methods=['GET'],
        collection_name='compjutahs',
        exclude_columns=['name'])

    expected = {
        'FlaskRestlessDatamodel': {
            'server_version': __version__,
            'serialize_naively': False,
        },
        'Computer': {
            'pk_name': 'id',
            'collection_name': 'compjutahs',
            'url_prefix': '/api',
            'attributes': {
                'id': 'integer',
                'owner_id': 'integer',
                'owner_name': 'unicode',
                'purchase_time': 'datetime',
                'vendor': 'unicode'
            },
            'relations': {
                'owner': {
                    'backref': 'computers',
                    'foreign_model': 'Person',
                    'local_column': 'owner_id',
                    'relation_type': 'MANYTOONE'
                },
                'peers': {
                    'foreign_model': 'Computer',
                    'is_proxy': True,
                    'relation_type': 'MANYTOONE'
                }
            },
            'properties': {},
            'methods': {}
        },
        'Person': {
            'pk_name': 'id',
            'collection_name': 'person',
            'url_prefix': '/api',
            'attributes': {
                'name': 'unicode'
            },
            'relations': {},
            'properties': {},
            'methods': {}
        }
    }

    client = client_maker(app)
    res = client.get('http://app/api/flask-restless-datamodel').json()
    assert res == expected


def test_inheritance(app, client_maker):
    db = SQLAlchemy(app)

    class Person(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        discriminator = db.Column(db.Unicode)
        __mapper_args__ = {'polymorphic_on': discriminator}

    class Engineer(Person):
        __mapper_args__ = {'polymorphic_identity': 'engineer'}
        id = db.Column(
            db.Integer, db.ForeignKey('person.id'), primary_key=True)
        primary_language = db.Column(db.Unicode)

    db.create_all()

    manager = flask_restless.APIManager(app, flask_sqlalchemy_db=db)
    manager.create_api(Person, methods=['GET'])
    manager.create_api(Engineer, methods=['GET'])
    data_model = DataModel(manager)
    manager.create_api(data_model, methods=['GET'])

    expected = {
        'FlaskRestlessDatamodel': {
            'server_version': __version__,
            'serialize_naively': False,
        },
        'Engineer': {
            'pk_name': 'id',
            'collection_name': 'engineer',
            'url_prefix': '/api',
            'polymorphic': {
                'identity': 'engineer',
                'parent': 'Person',
            },
            'attributes': {
                'id': 'integer',
                'primary_language': 'unicode',
                'discriminator': 'unicode'
            },
            'relations': {},
            'properties': {},
            'methods': {}
        },
        'Person': {
            'pk_name': 'id',
            'collection_name': 'person',
            'url_prefix': '/api',
            'polymorphic': {
                'on': 'discriminator',
                'identities': {
                    'engineer': 'Engineer'
                }
            },
            'attributes': {
                'id': 'integer',
                'discriminator': 'unicode'
            },
            'relations': {},
            'properties': {},
            'methods': {}
        }
    }

    client = client_maker(app)
    res = client.get('http://app/api/flask-restless-datamodel').json()
    assert res == expected


@pytest.fixture(scope='function')
def exposed_method_model_app(app):
    return _exposed_method_model_app(app)


@pytest.fixture(scope='function')
def exposed_method_model_app_with_commit(app):
    return _exposed_method_model_app(app, commit_before_return=True)


def _exposed_method_model_app(app, commit_before_return=False):
    db = SQLAlchemy(app)

    class Person(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        name = db.Column(db.Unicode, unique=True)
        birth_date = db.Column(db.Date)

        @property
        def id_to_text(self):
            return 'one'

        @property
        def settable_property(self):
            return self.name

        @settable_property.setter
        def settable_property(self, value):
            self.name = value

        def age_in_x_years_y_months(self,
                                    y_offset,
                                    m_offset=0,
                                    *args,
                                    **kwargs):
            dt = self.birth_date
            return dt.replace(
                year=dt.year + y_offset, month=dt.month + m_offset)

        def get_attrs_based_on_dict(self, args):
            return_dict = {}
            for key, val in args.items():
                if val:
                    return_dict[key] = getattr(self, key)
            return return_dict

        def what_does_this_func_even_do(self, person):
            assert isinstance(person, Person)
            return person

        def raise_an_error(self):
            raise Exception('Something happened')

        def create_person_but_dont_commit(self):
            person = Person(name='Some dude')
            db.session.add(person)
            db.session.flush()
            return person

    app.Person = Person
    db.create_all()

    db.session.add(Person(name='Jim Darkmagic', birth_date=date(2018, 1, 1)))
    db.session.commit()

    manager = flask_restless.APIManager(app, flask_sqlalchemy_db=db)
    manager.create_api(Person, methods=['GET'])
    data_model = DataModel(
        manager,
        include_model_functions=True,
        commit_on_method_return=commit_before_return)
    manager.create_api(data_model, methods=['GET'])
    return app


def test_exposed_methods(exposed_method_model_app, client_maker):
    client = client_maker(exposed_method_model_app)
    res = client.get('http://app/api/flask-restless-datamodel').json()

    expected = {
        'FlaskRestlessDatamodel': {
            'server_version': __version__,
            'serialize_naively': False,
        },
        'Person': {
            'pk_name': 'id',
            'collection_name': 'person',
            'url_prefix': '/api',
            'attributes': {
                'id': 'integer',
                'name': 'unicode',
                'birth_date': 'date',
            },
            'properties': {
                'id_to_text': False,
                'settable_property': True
            },
            'relations': {},
            'methods': {
                'age_in_x_years_y_months': {
                    'args': ['y_offset'],
                    'kwargs': ['m_offset'],
                    'argsvar': 'args',
                    'kwargsvar': 'kwargs'
                },
                'get_attrs_based_on_dict': {
                    'args': ['args'],
                    'kwargs': [],
                    'argsvar': None,
                    'kwargsvar': None
                },
                'what_does_this_func_even_do': {
                    'args': ['person'],
                    'kwargs': [],
                    'argsvar': None,
                    'kwargsvar': None
                },
                'raise_an_error': {
                    'args': [],
                    'kwargs': [],
                    'argsvar': None,
                    'kwargsvar': None
                },
                'create_person_but_dont_commit': {
                    'args': [],
                    'kwargs': [],
                    'argsvar': None,
                    'kwargsvar': None
                },
            }
        }
    }
    assert res == expected


def to_method_params(body, sr):
    return {'payload': sr.dumps(body)}


def test_call_exposed_method(exposed_method_model_app, client_maker):
    client = client_maker(exposed_method_model_app)
    sr = exposed_method_model_app.extensions['cereal']
    url = 'http://app/api/method/person/1/age_in_x_years_y_months'
    body = to_method_params({
        'args': [],
        'kwargs': {
            'y_offset': 10,
            'm_offset': 3
        }
    }, sr)
    res = sr.loads(client.post(url, json=body).json()['payload'])
    expected = date(2028, 4, 1)
    assert res == expected


def test_call_exposed_method_raises_an_error(exposed_method_model_app,
                                             client_maker):
    client = client_maker(exposed_method_model_app)
    url = 'http://app/api/method/person/1/raise_an_error'
    sr = exposed_method_model_app.extensions['cereal']
    body = to_method_params({'args': [], 'kwargs': {}}, sr)

    res = client.post(url, json=body)
    assert res.status_code == 500
    assert res.json() == {'message': 'Exception: Something happened'}


def test_call_exposed_method_with_model(exposed_method_model_app,
                                        client_maker):
    client = client_maker(exposed_method_model_app)
    url = 'http://app/api/method/person/1/what_does_this_func_even_do'
    sr = exposed_method_model_app.extensions['cereal']
    client_cereal = Cereal()

    class Person:
        id = 1

    client_cereal.register_class('Person', Person, lambda x: {'id': x.id},
                                 lambda x: x)

    body = to_method_params({
        'args': [],
        'kwargs': {
            'person': Person()
        }
    }, client_cereal)
    res = sr.loads(client.post(url, json=body).json()['payload'])
    assert res.name == 'Jim Darkmagic'


def run_transient_method(client, status, sr):
    url = 'http://app/api/method/person/1/create_person_but_dont_commit'
    body = to_method_params({'args': [], 'kwargs': {}}, sr)
    client_cereal = Cereal()
    client_cereal.register_class('Person', None, None, lambda x: x)

    res = client_cereal.loads(client.post(url, json=body).json()['payload'])
    assert res == {'id': 2, 'name': 'Some dude', 'birth_date': None}

    url = 'http://app/api/person/2'
    assert client.get(url).status_code == status


def test_it_doesnt_commit_transient_objects(exposed_method_model_app,
                                            client_maker):
    client = client_maker(exposed_method_model_app)
    sr = exposed_method_model_app.extensions['cereal']
    run_transient_method(client, 404, sr)


def test_it_commits_transient_objects(exposed_method_model_app_with_commit,
                                      client_maker):
    client = client_maker(exposed_method_model_app_with_commit)
    sr = exposed_method_model_app_with_commit.extensions['cereal']
    run_transient_method(client, 200, sr)


def test_it_can_identify_a_hybrid_property(app, client_maker):
    db = SQLAlchemy(app)

    class Person(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        first_name = db.Column(db.Unicode)
        last_name = db.Column(db.Unicode)

        @hybrid_property
        def name(self):
            return "{} {}".format(self.first_name, self.last_name)

    db.create_all()

    manager = flask_restless.APIManager(app, flask_sqlalchemy_db=db)
    manager.create_api(Person, methods=['GET'])
    data_model = DataModel(manager)
    manager.create_api(data_model, methods=['GET'])

    expected = {
        'FlaskRestlessDatamodel': {
            'server_version': __version__,
            'serialize_naively': False,
        },
        'Person': {
            'attributes': {
                'first_name': 'unicode',
                'id': 'integer',
                'last_name': 'unicode',
                'name': 'hybrid'
            },
            'collection_name': 'person',
            'url_prefix': '/api',
            'methods': {},
            'pk_name': 'id',
            'properties': {},
            'relations': {},
        }
    }

    client = client_maker(app)
    res = client.get('http://app/api/flask-restless-datamodel').json()
    assert res == expected


def test_it_can_identify_a_property(app, client_maker):
    db = SQLAlchemy(app)

    class Person(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        first_name = db.Column(db.Unicode)
        last_name = db.Column(db.Unicode)

        @hybrid_property
        def name(self):
            return "{} {}".format(self.first_name, self.last_name)

        @property
        def lower_name(self):
            return self.name.lower()

    db.create_all()

    manager = flask_restless.APIManager(app, flask_sqlalchemy_db=db)
    manager.create_api(Person, methods=['GET'])
    data_model = DataModel(manager)
    manager.create_api(data_model, methods=['GET'])

    expected = {
        'FlaskRestlessDatamodel': {
            'server_version': __version__,
            'serialize_naively': False,
        },
        'Person': {
            'attributes': {
                'first_name': 'unicode',
                'id': 'integer',
                'last_name': 'unicode',
                'name': 'hybrid',
            },
            'collection_name': 'person',
            'url_prefix': '/api',
            'methods': {},
            'properties': {
                'lower_name': False
            },
            'pk_name': 'id',
            'relations': {}
        }
    }

    client = client_maker(app)
    res = client.get('http://app/api/flask-restless-datamodel').json()
    assert res == expected


def test_it_can_get_a_property(exposed_method_model_app, client_maker):
    client = client_maker(exposed_method_model_app)
    url = 'http://app/api/property/person/1/id_to_text'
    sr = exposed_method_model_app.extensions['cereal']
    res = sr.loads(client.get(url).json()['payload'])
    assert res == 'one'


def test_it_can_set_a_property(exposed_method_model_app, client_maker):
    app = exposed_method_model_app
    client = client_maker(app)
    url = 'http://app/api/property/person/1/settable_property'
    sr = app.extensions['cereal']
    expected = 'new_value'

    body = sr.dumps(expected)
    res = client.post(url, json=body)
    person = app.Person.query.get(1)
    assert person.name == expected



import inspect

from flask_restless.helpers import get_related_association_proxy_model, primary_key_name
from sqlalchemy.ext.associationproxy import AssociationProxy
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.inspection import inspect as sqla_inspect
from sqlalchemy.orm.properties import ColumnProperty, RelationshipProperty

from .helpers import object_property, register_serializer, run_object_method

INCLUDE_INTERNAL = 'include_model_internal_functions'
COMMIT_ON_RETURN = 'commit_on_method_return'
EXPOSE_PROPERTY = 'expose_property'


def clean(columns):
    return columns or []


def get_is_valid_validator(included, excluded):
    def is_valid(column):
        column = column.split('.')[-1]
        valid_excl = True
        valid_incl = True
        if excluded:
            valid_excl = column not in excluded
        if included:
            valid_incl = column in included
        return valid_excl and valid_incl

    return is_valid


def is_polymorphic(model, check_var):
    has_mapper_args = hasattr(model, '__mapper_args__')
    if has_mapper_args and check_var in model.__mapper_args__:
        return True
    return False


class DataModelRenderer:
    def __init__(self, app, db, options):
        self.app = app
        self.options = options

    def render(self, model, config):
        klass = ClassDefinitionRenderer(self.app, self.options, model, config)
        methods = MethodDefinitionRenderer(self.options, model, config)
        model_render = klass.render()
        model_render['methods'] = methods.render()
        return model_render

    def render_polymorphic(self, model, identities):
        polymorphic_info = {}
        if is_polymorphic(model, 'polymorphic_on'):
            mapper_args = model.__mapper_args__
            on = mapper_args['polymorphic_on']
            if not isinstance(on, str):
                on = on.key
            polymorphic_info['on'] = on
            polymorphic_info['identities'] = identities
        if is_polymorphic(model, 'polymorphic_identity'):
            mapper_args = model.__mapper_args__
            for kls in model.__bases__:
                if is_polymorphic(kls, 'polymorphic_on'):
                    polymorphic_info['parent'] = kls.__name__
                    polymorphic_info['identity'] = mapper_args[
                        'polymorphic_identity']
        return polymorphic_info


class ClassDefinitionRenderer:
    def __init__(self, app, options, model, config):
        self.app = app
        self.model = model
        self.options = options
        self.config = config
        self.is_valid = get_is_valid_validator(
            clean(config.view.include_columns),
            clean(config.view.exclude_columns))

    def render(self):
        view = self.config.view
        collection_name = self.config.collection_name

        attribute_dict = self.render_attributes()
        foreign_keys = self.render_relations()
        properties = {}
        if self.options.get(EXPOSE_PROPERTY, True):
            properties = self.render_properties()
        attribute_dict.update(self.render_hybrid_properties())
        self.render_association_proxies(attribute_dict, foreign_keys)

        with self.app.app_context():
            pk_name = primary_key_name(self.model)

        cr = self.app.extensions['cereal']
        register_serializer(self.model, pk_name, view.serialize,
                            view.deserialize, cr)

        return {
            'pk_name': pk_name,
            'collection_name': collection_name,
            'url_prefix': self.config.blueprint.url_prefix,
            'attributes': attribute_dict,
            'relations': foreign_keys,
            'properties': properties,
        }

    def render_attributes(self):
        attribute_dict = {}
        for column in sqla_inspect(self.model).columns:
            if self.is_valid(column.name):
                ctype = column.type.__class__.__name__.lower()
                attribute_dict[column.name] = ctype
        return attribute_dict

    def render_relations(self):
        foreign_keys = {}
        for rel in sqla_inspect(self.model).relationships:
            if self.is_valid(str(rel.key)):
                direction = rel.direction.name
                if rel.direction.name == 'ONETOMANY' and not rel.uselist:
                    direction = 'ONETOONE'
                foreign_keys[rel.key] = {
                    'foreign_model': rel.mapper.class_.__name__,
                    'relation_type': direction,
                    'backref': rel.back_populates,
                }
                if rel.direction.name == 'MANYTOONE':
                    local_id = list(rel.local_columns)[0].key
                    foreign_keys[rel.key]['local_column'] = local_id
        return foreign_keys

    def render_properties(self):
        attribute_dict = {}
        properties = [(a, getattr(self.model, a).fset is not None)
                      for a in dir(self.model)
                      if isinstance(getattr(self.model, a), property)]
        for attribute, settable in properties:
            if self.is_valid(attribute):
                self.add_property_endpoint(attribute)
                attribute_dict[attribute] = settable

        return attribute_dict

    def render_hybrid_properties(self):
        attribute_dict = {}
        hybrid_properties = [
            a for a in sqla_inspect(self.model).all_orm_descriptors
            if isinstance(a, hybrid_property)
        ]
        for attribute in hybrid_properties:
            name = attribute.__name__
            if self.is_valid(name):
                attribute_dict[name] = 'hybrid'
        return attribute_dict

    def render_association_proxies(self, attribute_dict, foreign_keys):
        proxies = {}
        for k in list(self.model.__dict__.keys()):
            v = self.model.__dict__[k]
            is_proxy = isinstance(v, AssociationProxy)
            # keep the proxies where the remote attr has a property,
            # as we need this property to identify the remote class
            # but not all cases have it.
            # v == v.__get__(None, model), but we do this to bind the model to
            # the remote_attr and from then on it's usable for further inspection
            if is_proxy and hasattr(
                    v.__get__(None, self.model).remote_attr, 'property'):
                proxies[k] = v.__get__(None, self.model)

        for name, attr in proxies.items():
            # check if the remote attr is a relation (for example, an association
            # table) or if it's an attribute
            if isinstance(attr.remote_attr.property, RelationshipProperty):
                # use the helper function from flask restless to identify the
                # remote class
                remote_class = get_related_association_proxy_model(attr)
                foreign_keys[name] = {
                    'foreign_model': remote_class.__name__,
                    'relation_type': 'MANYTOONE'
                                     if attr.scalar else 'ONETOMANY',
                    'is_proxy': True
                }
            elif isinstance(attr.remote_attr.property, ColumnProperty):
                # The columns of remote attr will always be 1 element in size
                # as the columns is refering to itself (i.e. the remote attr)
                column = attr.remote_attr.property.columns[0]
                attribute_dict[name] = column.type.__class__.__name__.lower()

    def add_property_endpoint(self, property_name):
        fmt = '/property/{0}/<instid>/{1}'
        endpoint = fmt.format(self.config.collection_name, property_name)
        self.config.blueprint.add_url_rule(
            endpoint,
            methods=['GET', 'POST'],
            defaults={
                'model': self.model,
                'property_name': property_name
            },
            view_func=object_property)


class MethodDefinitionRenderer:
    def __init__(self, options, model, config):
        self.options = options
        self.model = model
        self.config = config

    def render(self):
        methods = self.compile_method_list()
        self.add_method_endpoints(methods)
        return methods

    def compile_method_list(self):
        methods = {}
        include_internal = self.options.get(INCLUDE_INTERNAL, False)
        for name, fn in inspect.getmembers(
                self.model, predicate=inspect.isfunction):
            if name.startswith('__'):
                continue
            if name.startswith('_') and not include_internal:
                continue

            spec = inspect.signature(fn)
            required = []
            optional = []
            argsvar = None
            kwargsvar = None
            for param_name, param in spec.parameters.items():
                if param_name == 'self':
                    continue
                if param.kind == param.VAR_KEYWORD:
                    kwargsvar = param_name
                elif param.kind == param.VAR_POSITIONAL:
                    argsvar = param_name
                elif param.default == param.empty:
                    required.append(param_name)
                else:
                    optional.append(param_name)

            methods[name] = {
                'args': required,
                'kwargs': optional,
                'argsvar': argsvar,
                'kwargsvar': kwargsvar,
            }
        return methods

    def add_method_endpoints(self, methods):
        commit_on_return = self.options.get(COMMIT_ON_RETURN, False)
        collection_name = self.config.collection_name
        for method in methods.keys():
            fmt = '/method/{0}/<instid>/{1}'
            instance_endpoint = fmt.format(collection_name, method)
            self.config.blueprint.add_url_rule(
                instance_endpoint,
                methods=['POST'],
                defaults={
                    'function_name': method,
                    'model': self.model,
                    'commit_on_return': commit_on_return,
                },
                view_func=run_object_method)



import json
from collections import namedtuple

import flask
from sqlalchemy.orm.session import Session

ModelConfiguration = namedtuple('ModelConfiguration',
                                'collection_name view blueprint')


def abort(msg):
    resp = flask.jsonify(message=msg)
    resp.status_code = 500
    flask.abort(resp)


def cr():
    return flask.current_app.extensions['cereal']


def register_serializer(model, pk_name, serialize, deserialize, cr):
    def load_model(value):
        pkval = value.get(pk_name)
        if pkval:
            return model.query.filter_by(**{pk_name: pkval}).one_or_none()
        return deserialize(value)

    def serialize_model(value):
        return serialize(value)

    cr.register_class(model.__name__, model, serialize_model, load_model)


def run_object_method(instid, function_name, model, commit_on_return):
    instance = model.query.get(instid)
    if not instance:
        return {}
    params = cr().loads(flask.request.get_json()['payload'])
    try:
        result = getattr(instance, function_name)(*params['args'],
                                                  **params['kwargs'])
        result = json.dumps({'payload': cr().dumps(result)})
    except Exception as e:
        msg = '{}: {}'.format(e.__class__.__name__, str(e))
        abort(msg)

    if commit_on_return:
        try:
            session = Session.object_session(instance)
            session.commit()
        except Exception:
            pass

    return result


def object_property(instid, model, property_name):
    if flask.request.method == 'GET':
        return get_object_property(instid, model, property_name)
    else:
        return set_object_property(instid, model, property_name)


def get_object_property(instid, model, property_name):
    instance = model.query.get(instid)
    if not instance:
        return {}
    result = getattr(instance, property_name)
    return json.dumps({'payload': cr().dumps(result)})


def set_object_property(instid, model, property_name):
    instance = model.query.get(instid)
    if not instance:
        return {}

    value = cr().loads(flask.request.get_json())
    try:
        setattr(instance, property_name, value)
        session = Session.object_session(instance)
        session.commit()
    except Exception as e:
        abort("Could not set property: {}".format(e))
    return json.dumps({'message': 'success'})
